from math import atan2,sqrt
import numpy as np

#Handle rand_py import
from . import rand_py as rand

"""
Inputs:
   p0, p1, p2: 3x1 vector
   k1, k2: 3x1 vector w/ norm = 1
Outputs:
   theta1, theta2: 1x1 angle in rads
"""

#Setup subproblem
def sp2E_setup(p0, p1, p2, k1, k2):
   #Clear all matrices to pass by ref
   p0 *= 0
   p1 *= 0
   p2 *= 0
   k1 *= 0
   k2 *= 0

   #Then pass in new values
   p0 += rand.rand_vec()
   p1 += rand.rand_vec()
   k1 += rand.rand_normal_vec()
   k2 += rand.rand_normal_vec()

   theta1 =  rand.rand_angle()
   theta2 = rand.rand_angle()

   #Calculate p2 using rand angles and above 
   #@ is index wise multiplication
   p2 += rand.rot(k2, -theta2) @ (p0 + rand.rot(k1, theta1) @ p1)


#Run subproblem
def sp2E_run(p0, p1, p2, k1, k2):
   KxP1 = np.cross(k1, p1)
   KxP2 = np.cross(k2, p2)
   
   #np.block appends two arrays together
   A1 = np.block([KxP1, -np.cross(k1, KxP1)])
   A2 = np.block([KxP2, -np.cross(k2, KxP2)])

   #Append A1, A2 then reshape into desired shape
   A = np.block([A1, -A2])
   A = np.reshape(A, (4,3))

   p = -k1*np.dot(k1, p1) + k2*np.dot(k2, p2) - p0
   
   radius_1_sq = np.dot(KxP1, KxP1)
   radius_2_sq = np.dot(KxP2, KxP2)

   alpha = radius_1_sq/(radius_1_sq+radius_2_sq)
   beta  = radius_2_sq/(radius_1_sq+radius_2_sq) 

   #reshape vector to matrix w/ opposite dimensions
   M_inv = np.eye(3) + (k1 * np.reshape(k1, (3,1)))*(alpha/(1.0-alpha)) 

   AAT_inv = 1.0/(radius_1_sq+radius_2_sq)*(M_inv + M_inv@np.array([k2]).T*k2 @ \
             M_inv*beta/(1.0-k2@M_inv@np.array([k2]).T*beta))

   x_ls = np.dot(A@AAT_inv,p)

   n_sym = np.reshape(np.cross(k1, k2), (3,1))
   pinv_A1 = A1/radius_1_sq
   pinv_A2 = A2/radius_2_sq

   A_perp_tilde = np.reshape(np.block([pinv_A1, pinv_A2]), (4, 3)) @ n_sym

   num = (pow(np.linalg.norm(x_ls[2:4], 2), 2) - 1.0)*pow(np.linalg.norm(A_perp_tilde[0:2], 2), 2) - \
         (pow(np.linalg.norm(x_ls[0:2], 2), 2) - 1.0)*pow(np.linalg.norm(A_perp_tilde[2:4], 2), 2)
   #den was in form [[den]], so we converted to integer through indexing
   den = 2*(np.reshape(x_ls[0:2], (1,2)) @ A_perp_tilde[0:2] * pow(np.linalg.norm(A_perp_tilde[2:4], 2), 2) - \
         np.reshape(x_ls[2:4], (1,2)) @ A_perp_tilde[2:4] * pow(np.linalg.norm(A_perp_tilde[0:2], 2), 2))[0][0]
   
   xi = num/den
   #We want sc as vector/list, so we flatten both arrays
   sc = x_ls.flatten() + xi*A_perp_tilde.flatten()

   #Commented out for optimization
   #theta1 = atan2(sc[0], sc[1])
   #theta2 = atan2(sc[2], sc[3])

   return [atan2(sc[0], sc[1]), atan2(sc[2], sc[3])]


#Error calculation code
def sp2E_error(p0, p1, p2, k1, k2, theta1, theta2):
   return np.linalg.norm(p0 + rand.rot(k1, theta1)@p1 - rand.rot(k2, theta2)@p2, 2)



#Test Code
if __name__ == "__main__":
   print("Starting arrays \r\n")
   p0 = np.array([1., 2., 3.])
   p1 = np.array([1., 2., 3.])
   p2 = np.array([1., 2., 3.])
   k1 = np.array([1., 2., 3.])
   k2 = np.array([1., 2., 3.])
   
   #Setup problem
   sp2E_setup(p0, p1, p2, k1, k2)
   print("After setup \r\n")
   np.set_printoptions(precision = 20)
   print("p0={}\r\np1={}\r\nk1={}\r\nk2={}\r\n\np2={}\r\n".format(p0, p1, k1, k2, p2))

   t1, t2 = sp2E_run(p0, p1, p2, k1, k2) #Save values

   #Printing out results
   print("Results:\r\ntheta1:\n{}\r\ntheta2:\n{}\r\n".format(t1, t2))
   print("Error:", sp2E_error(p0, p1, p2, k1, k2, t1, t2), "\r\n")


from math import atan2
import numpy as np

#Handle rand_py, sp1_lib imports
from . import rand_py as rand
from . import sp1_lib as sp1

"""
Inputs:
   p0, p1, p2, p3: 3x1 vector
   k1, k2, k3: 3x1 vector w/ norm = 1
Outputs:
   theta1, theta2, theta3: 1xN angle in rads where N = num sols
"""


#Setup subproblem
def sp5_setup(p0, p1, p2, p3, k1, k2, k3):
   #print("Setting up\r\n")
   #Clear all matrices to pass by ref
   p0 *= 0
   p1 *= 0
   p2 *= 0
   p3 *= 0
   k1 *= 0
   k2 *= 0
   k3 *= 0

   #Then pass in new values
   p1 += rand.rand_vec()
   p2 += rand.rand_vec()
   p3 += rand.rand_vec()

   k1 += rand.rand_normal_vec()
   k2 += rand.rand_normal_vec()
   k3 += rand.rand_normal_vec()

   theta1 = rand.rand_angle()
   theta2 = rand.rand_angle()
   theta3 = rand.rand_angle()

   #print("theta1=", theta1, "\r\ntheta2=", theta2, "\r\ntheta3=", theta3, "\r\n")

   p0 += -(rand.rot(k1, theta1) @ p1 - rand.rot(k2, theta2) @ (p2 + rand.rot(k3, theta3) @ p3))


#Represents polynomials as coeffecient vectors
def cone_polynomials(p0_i, k_i, p_i, p_i_s, k2):
   #||A x + p_S - H k_2||^2 = 
   #-H^2 + P(H) +- sqrt(R(H))
   #Reperesent polynomials P_i, R_i as coefficient vectors
   #(Highest powers of H first)
   kiXk2 = np.cross(k_i, k2)
   kiXkiXk2 = np.cross(k_i, kiXk2)
   norm_kiXk2_sq = np.dot(kiXk2, kiXk2)

   kiXpi = np.cross(k_i, p_i)
   norm_kiXpi_sq = np.dot(kiXpi, kiXpi)

   delta = np.dot(k2 ,p_i_s)
   alpha = kiXkiXk2 @ np.reshape(p0_i, (3, 1)) / norm_kiXk2_sq
   beta =  kiXk2 @ np.reshape(p0_i, (3, 1)) / norm_kiXk2_sq

   P_const = norm_kiXpi_sq + np.dot(p_i_s, p_i_s) + 2 * alpha * delta
   #P = np.array([-2*alpha, P_const])
   R = np.array([-1, 2*delta, -pow(delta, 2)]) #-(H-delta_i)^2
   R[len(R) - 1] = R[len(R) - 1] + norm_kiXpi_sq*norm_kiXk2_sq #||A_i' k_2||^2 - (H-delta_i)^2
   return np.array([-2.0*alpha, P_const]), pow(2.0*beta, 2) * R


#Run subproblem
def sp5_run(p0, p1, p2, p3, k1, k2, k3):
   i_soln = 0

   p1_s = p0 + np.reshape(k1, (3, 1)) @ np.reshape(k1, (1, 3)) @ p1
   p3_s = p2 + np.reshape(k3, (3, 1)) @ np.reshape(k3, (1, 3)) @ p3
   delta1 = np.dot(k2, p1_s)
   delta3 = np.dot(k2, p3_s)   

   P_1, R_1 = cone_polynomials(p0, k1, p1, p1_s, k2)
   P_3, R_3 = cone_polynomials(p2, k3, p3, p3_s, k2)

   #Now solve the quadratic for H   
   P_13 = P_1 - P_3

   #Use 1D version of matrix
   P_13_sq = np.convolve(P_13[:, 0], P_13[:, 0])

   RHS = R_3 - R_1 - P_13_sq
   EQN = np.convolve(RHS, RHS)-4.0 * np.convolve(P_13_sq, R_1)

   all_roots = np.reshape(np.roots(EQN), (np.roots(EQN).size, 1))
   H_vec = all_roots[np.real(all_roots) == all_roots] #Find only those w/ imaginary part == 0

   H_vec = H_vec.real # for coder, removes non-existant imaginary part
   
   KxP1 = np.cross(k1, p1)
   KxP3 = np.cross(k3, p3)

   #np.concatenate appends two arrays together
   A1 = np.concatenate(([KxP1], [-np.cross(k1, KxP1)]), axis = 0).T
   A3 = np.concatenate(([KxP3], [-np.cross(k3, KxP3)]), axis = 0).T

   signs = [[1, 1, -1, -1], [1, -1, 1, -1]]
   J = np.array([[0, 1], [-1, 0]])

   #Setup return variables
   theta1 = []
   theta2 = []
   theta3 = []
   for i_H in range(len(H_vec)):
      H = H_vec[i_H]

      const_1 = k2 @ A1 * (H-delta1)
      const_3 = k2 @ A3 * (H-delta3)
      
      pm_1 = k2 @ A1 @ J * np.emath.sqrt(np.linalg.norm(k2 @ A1, 2)*np.linalg.norm(k2 @ A1, 2) - (H-delta1)*(H-delta1))
      pm_3 = k2 @ A3 @ J * np.emath.sqrt(np.linalg.norm(k2 @ A3, 2)*np.linalg.norm(k2 @ A3, 2) - (H-delta3)*(H-delta3))

      for i_sign in range(4):
         sign_1 = signs[0][i_sign]
         sign_3 = signs[1][i_sign]

         sc1 = const_1 + sign_1 * pm_1
         sc1 = sc1 / pow(np.linalg.norm(k2 @ A1, 2), 2)

         sc3 = const_3 + sign_3 * pm_3
         sc3 = sc3 / pow(np.linalg.norm(k2 @ A3, 2), 2)

         v1 = A1 @ sc1 + p1_s
         v3 = A3 @ sc3 + p3_s

         if abs(np.linalg.norm(v1-H*k2, 2) - np.linalg.norm(v3-H*k2, 2)) < 1E-6:
            i_soln = 1 + i_soln
            theta1.append(atan2(sc1[0], sc1[1]))
            theta2.append(sp1.sp1_run(v3, v1, k2)[0])
            theta3.append(atan2(sc3[0], sc3[1]))
   return theta1, theta2, theta3


def sp5_error(p0, p1, p2, p3, k1, k2, k3, theta1, theta2, theta3):
   e = 0.0 #Sum variable
   #Perform calculations on each index
   for i in range(len(theta1)):
      e_i = np.linalg.norm(p0 + rand.rot(k1, theta1[i]) @ p1 - rand.rot(k2, theta2[i]) @ (p2 + rand.rot(k3, theta3[i]) @ p3), 2)
      e += e_i
   return e



#Test Code
if __name__ == "__main__":
   print("Starting arrays \r\n")
   
   p0 = np.array([1., 2., 3.])
   p1 = np.array([1., 2., 3.])
   p2 = np.array([1., 2., 3.])
   p3 = np.array([1., 2., 3.])
   k1 = np.array([1., 2., 3.])
   k2 = np.array([1., 2., 3.])
   k3 = np.array([1., 2., 3.])
   
   #Setup problem
   sp5_setup(p0, p1, p2, p3, k1, k2, k3)
   np.set_printoptions(precision = 20)
   
   print(("p1={}\r\np2={}\r\np3={}\r\nk1={}\r\nk2={}\r\nk3={}\r\n\np0={}\r\n".format(p1, p2, p3, k1, k2, k3, p0)))
   #MATLAB friendly output is below
   #print(" ".join(("p1=" + str(p1) + "!p2=" + str(p2) + "!p3=" + str(p3) + "!k1=" + str(k1) + "!k2=" + str(k2) + "!k3=" + str(k3) + "!p0=" + str(p0)).split()).replace(" ", ";").replace("!", "\n").replace(";]", "]"), "\r\n")
   
   t1, t2, t3 = sp5_run(p0, p1, p2, p3, k1, k2, k3) #Save values

   #Printing out results
   print("Results:\r\ntheta1:\n{}\r\ntheta2:\n{}\r\ntheta3:\n{}\r\n".format(t1, t2, t3))
   print("Error:", sp5_error(p0, p1, p2, p3, k1, k2, k3, t1, t2, t3), "\r\n")


import numpy as np #Used for matrix operations
import math        #Used for pi
import random as r

def hat(k):
   return np.array([[0, -k[2], k[1]], [k[2], 0, -k[0]], [-k[1], k[0], 0]])

#Create a random angle
def rand_angle():
   return round(r.random(), 15)*2*math.pi-math.pi

#Create random matrix of size 3 x size (rows, cols)
def rand_mat(size = 1): #Make list of lists, then pass to numpy
   return np.array([[round(r.random(), 15) for x in range(3)] for y in range(size)])

#Create random vector of size 3 x N
def rand_vec(): #Make list of lists, then pass to numpy
   return np.array([round(r.random(), 15) for y in range(3)])

#Create random normal vector
def rand_normal_vec():
   vec = rand_vec()
   return vec/np.linalg.norm(vec, 2)

#Create random normal matrix
def rand_normal_mat(size = 1):
   #vec = rand_vec()
   #return np.array([[rand_vec()/] for y in range(0, 3)])
   lst = np.zeros((size, 3))
   for x in range(size):
      lst[x] = [round(r.random(), 15) for y in range(3)]
      lst[x] = lst[x] / np.linalg.norm(lst[x], 2)
   
   return lst

#Create rand perp vec from input numpi.array
def rand_perp_normal_vec(inp):
   randCross = rand_vec(1)
   randCross = np.cross(randCross, inp)
   return randCross/np.linalg.norm(randCross, 2)

def rot(k, theta):
   k = k/np.linalg.norm(k,2)
   return np.eye(3) + math.sin(theta)*hat(k)+(1-math.cos(theta))*hat(k)@hat(k)
   


from math import atan2,sqrt
import numpy as np

#Handle rand_py import
from . import rand_py as rand


"""
Inputs:
   p0, p1, p2, p3: 3x1 vector
   k1, k2, k3: 3x1 vector w/ norm = 1
Outputs:
   theta1: 1xN angle vec in rad, N = num sols
   is_LS: 1x1 bool

This problem will be ill-posed if p is parallel to k or h
"""

#Setup for non-least squares version
def sp4_setup(p, k, h):
   #Clear all matrices to pass by ref
   p *= 0
   k *= 0
   h *= 0

   #Then pass in new values
   p += rand.rand_vec()
   k += rand.rand_normal_vec()
   h += rand.rand_normal_vec()
   
   theta = rand.rand_angle()
   #print("theta=", theta, "\r\n") #If seeing theta is desired, uncomment this.

   #Return d
   return h @ rand.rot(k, theta) @ p


def sp4_setup_LS(p, k, h):
   #Clear all matrices to pass by ref
   p *= 0
   k *= 0
   h *= 0

   #Then pass in new values
   p += rand.rand_vec()
   k += rand.rand_normal_vec()
   h += rand.rand_normal_vec()

   #Return d
   return rand.r.random() #rand 0-1


def sp4_run(p, k, h, d):
   A11 = np.cross(k,p)
   A1 = np.concatenate(([A11], [-np.cross(k,A11)]), axis = 0)
   A = h @ A1.T

   b = d - h @ np.reshape(k, (3, 1)) @ (k @ np.reshape(p, (3, 1)))
   norm_A2 = A @ A
   x_ls = A1 @ (h * b)

   #If a non-least squares version
   if norm_A2 > pow(b, 2):
      xi = sqrt(norm_A2-pow(b, 2))
      A_perp_tilde = np.block([A[1], -A[0]])
      
      sc_1 = x_ls + xi * A_perp_tilde
      sc_2 = x_ls - xi * A_perp_tilde

      #theta = [atan2(sc_1[0], sc_1[1]), atan2(sc_2[0], sc_2[1])]
      return np.array([atan2(sc_1[0], sc_1[1]), atan2(sc_2[0], sc_2[1])]), False
   
   #Otherwise problem is least squares
   else:
      #theta = atan2(x_ls[0], x_ls[1])
      return atan2(x_ls[0], x_ls[1]), True

def sp4_error(p, k, h, d, theta):
   #If least-squares:     theta is a float
   #If not least-squares: theta is a np.array
   if type(theta) != float:
      e = 0.0 #Sum variable
      #Perform calculations on each index
      for i in range(len(theta)):
         e_i = np.linalg.norm(h @ rand.rot(k, theta[i]) @ p) - d
         e += e_i
   #If least-squares, then no summation needed
   else:
      e = np.linalg.norm(h @ rand.rot(k, theta) @ p) - d
   return e



#Test Code
if __name__ == "__main__":
   print("Starting arrays \r\n")
   p = np.array([1., 2., 3.])
   k = np.array([1., 2., 3.])
   h = np.array([1., 2., 3.])

   #Setup problem
   print("After setup: \r\n")
   d = sp4_setup_LS(p, k, h)
   np.set_printoptions(precision = 20)
   print("p = {}\r\nk = {}\r\nh  = {}\r\nd  = {}\r\n".format(p, k, h, d))

   theta, is_LS = sp4_run(p, k, h, d) #Save values

   #Printing out results
   print("Results:\r\ntheta: {}\r\nis_LS: {}\r".format(theta, is_LS))
   print("Error:", sp4_error(p, k, h, d, theta), "\r\n")